//    Message structure parser
//    Different message types (see RFC-1521)
//
//
//---------------------------------------------------------------------------
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <mem.h>
#include <sys/types.h>
#include <regex.h>

#include "memory.h"
#include "defs.h"
#include "eurist.h"

//---------------------------------------------------------------------------
#define	CharSet				0x0001
#define	MultiPart			0x0002
#define	Mime				0x0004
#define	UnKnown				0x0008
#define	ConTyp				0x0010
#define	ConTrEn	    			0x0020
#define	Empty				0x0040
#define	BadUnKnown			0x0080
#define	WaitingForBoundary		0x0100
#define	AlreadyDeveloped                0x0200
#define Digest                          0x0400
//#define	SetPlain                        0x0400


int searchfor(char * mask, char	* src)
{
int i,len;

len=strlen(mask);

for (i=0;((src[i]!=0) && (src[i]!='\n') && (src[i]!='\r'));i++)
   	if (strnicmp(mask,src+i,len)==0)
		return(i+len);
return(-1);
}

void unquote(char *src)
{
int i;
char *p;

if (src[0]=='\"')
   {
   for (i=1; ((src[i]!=0) && (src[i]!='\n') && (src[i]!='\r') && (src[i]!='\"')); i++)
       src[i-1]=src[i];
   src[i-1]=0;
   }
else
   {
   if ((p=strchr(src,'\n')) != NULL)
   		*p=0;
   if ((p=strchr(src,'\r')) != NULL)
   		*p=0;
   }
}


inline int IsSpace(char c)
{
if ((c==' ') || (c=='\t'))
	return(1);
else
	return(0);
}

// Returns -1 if they are really different
// 0 if equal
// positive integer - index in src to point just after mask

int lstrcmpi(const char * mask,const char * src)
{
int i,j;

if (tolower(*mask)!=tolower(*src))
	return(-1);

for (i=j=1;mask[i];i++,j++)
    {
    if (tolower(mask[i])==tolower(src[j]))
    	continue;

    if (!IsSpace(mask[--i]))
       return(-1);

    if (!IsSpace(src[j]))
    	return(-1);
    }

while (IsSpace(src[j]))
	j++;
return(src[j] ? j : 0);
}

int Lstrlen(char *p)
{
register char *s=p;

while(*s)
   if (*(s++)=='\n')
      break;

return((s-p) ? (s-p) : 1);
}

long Memory::DevelopMail(FILE * output, char * header)
{
unsigned int	flags=Empty;
int				pos,length,n,m;
int				charsetpos=-1;
int                             plainpos=-1;
char *				p;
char				bound[LINE_SIZE];
char				charset[LINE_SIZE];


for (pos=0;(p=Pointer2Line(0,pos))!=NULL;pos+=length)
	{
        length=Lstrlen(p);

	if ((*p=='\n') || ((*p=='\r') && (*(p+1)=='\n')))
    	   break;

        if ((header!=NULL) && (*header!='\n') && (strncmp(p,header,strlen(header)-1)==0))
           {
           flags|=AlreadyDeveloped;
           break;
           }

    	if (flags&WaitingForBoundary)
    	   {
	   n=0;
           goto WaitedBound;
           }

	flags&=~Empty;

    if (lstrcmpi("mime-version:",p)>=0)
    	{
        flags|=Mime;
        continue;
        }

    if (lstrcmpi("content-transfer-encoding:",p)>=0)
    	{
        flags|=ConTrEn;
        continue;
        }

    if ((n=lstrcmpi("content-type:",p))>=0)
    	{
        flags|=ConTyp;
        if ( (m=lstrcmpi("multipart/", p+n))>=0	)
           {
           if (lstrcmpi("multipart/digest", p+n)>=0)
              flags|=Digest;

// else the message is supposed to be multipart/mixed - primary type

           n+=m;

WaitedBound:
	   if ((m=searchfor("boundary=",p+n))>=0)
	      {
	      n+=m;
              bound[LINE_SIZE-1]=0;
              bound[0]=bound[1]='-';
	      strncpy(bound+2,p+n,LINE_SIZE-3);
	      unquote(bound+2);
	      flags|=MultiPart;
	      flags&=~WaitingForBoundary;
	      continue;
	      }
           else
	      {
	      if (flags&WaitingForBoundary)
	         {
	         flags|=BadUnKnown;
		 break;
		 }
              else
                 {
	         flags|=WaitingForBoundary;
		 continue;
		 }
              }
	   }

	else
		{
	    if ( (m=lstrcmpi("text", p+n))<0 )
               {
	       flags|=UnKnown;
	       break;
	       }
	    n+=m;
            if ( (p[n]=='\n') || (p[n]=='\r') )
               plainpos=pos+n;
	    if ( (m=searchfor("charset=", p+n))<0 )
	    	{
		charsetpos=pos+length-1;
                if (p[length-2]=='\r')
                   charsetpos--;
		if (((p=Pointer2Line(0,pos+length))!=NULL) &&	((n=searchfor("charset=", p))>=0 ))
		    {
		    pos=pos+length;

                    length=Lstrlen(p);

		    goto chr;
		    }
		}
	    else
	    	{
		n+=m;
chr:
		strncpy(charset,p+n,LINE_SIZE-1-n);
		unquote(charset);
		flags|=CharSet;
   		charsetpos=pos+n;
		}
	    }
	}
    }

// Header is scanned
if (flags& (WaitingForBoundary | BadUnKnown))
	MyError(LoMevin);

if  (((flags&AlreadyDeveloped) == 0) && (header!=NULL) && (*header!='\n'))                       // Custom header
    fwrite(header,1,strlen(header),output);

if (flags& (WaitingForBoundary | UnKnown | Empty | BadUnKnown | AlreadyDeveloped))
   {
   for (pos=0;(p=Pointer2Line(0,pos))!=NULL;pos+=length)
       {
       length=Lstrlen(p);
       fwrite(p,1,length,output);
       }
   return(pos);
   }

if ((flags&MultiPart)==0)
    {
    eurist eu=&data;

int itmp=0;
    while ((p=Pointer2Line(0,pos))!=NULL)
	{
	pos+=Lstrlen(p);
	eu+=(unsigned char *)p;
        }

    if (flags&CharSet)
    	{
	regmatch_t rm;

	if (regexec(&charset_regex,charset,0,&rm,0))
        	{
                if (data.charset_reliability==RELIABLE)
			strcpy(charset,eu.charset());
                }
	else
                {
                if (data.charset_reliability==UNRELIABLE)
			strcpy(charset,eu.charset());
                }
	}
    else
	strcpy(charset,eu.charset());
    }
else
    strcpy(charset,data.charset_name[0]);

for (pos=0;(p=Pointer2Line(0,pos))!=NULL;pos+=Lstrlen(p))
    {
    if ((*p=='\n') || ((*p=='\r') && (*(p+1)=='\n')))
       break;

    for (m=n=0;(p[n]!='\n')&&(p[n]!=0);n++)		//m is a flag for 8bit symbols in current line
	{
        if (n+pos==plainpos)
           {
           fputs("/plain",output);
           plainpos=-1;
           }

	if (n+pos==charsetpos)
           {
	   m=0;
	   break;
	   }
	if ( ((p[n]>='!')&&(p[n]<='<')) || ((p[n]>='>')&&(p[n]<='^'))   ||
             ((p[n]>='`')&&(p[n]<='~')) || (p[n]=='\n') || (p[n]=='\r') ||
             (p[n]=='\t') || (((p[n]==' ')||(p[n]=='_')||(p[n]=='=')) && (m==0)) )
	   fputc(p[n], output);
	else
	   {
	   if (!m)
              fprintf(output,"=?%s?Q?",charset);
           m=1;
           fprintf(output,"=%02X",(unsigned char)p[n]);
	   }
	}
    if (m)
       	fputs("?=",output);

    if (n+pos==plainpos)
       {
       fputs("/plain",output);
       plainpos=-1;
       }
    if (n+pos==charsetpos)
       {
       if (flags&CharSet)
	   fprintf(output, charset);
       else
           fprintf(output,"; charset=%s",charset);
       fputc('\n',output);
       continue;
       }

    if ((p[n]) || (n==0))
       fputc(p[n],output);
    }

if (!(flags&Mime))
 	fprintf(output,"MIME-Version: 1.0\n");

if (!(flags&ConTrEn))
	fprintf(output,"Content-Transfer-Encoding: 8bit\n");

if (!(flags&ConTyp))
   	fprintf(output,"Content-Type: text/plain; charset=%s\n",charset);

if (flags&MultiPart)
	{
	while ((p=Pointer2Line(0,pos))!=NULL)
              {
              pos+=(length=Lstrlen(p));
	      fwrite(p,1,length,output);
	      if (memcmp(p,bound,strlen(bound)))
                 continue;
              p=Pointer2Line(0,pos);
	      if (flags&Digest)
	         {
	         if ((p!=NULL) && (*p=='\r'))
	            {
	            pos++;
	            p++;
	            fputc('\r',output);
	            }
                 if ((p!=NULL) && (*p=='\n'))
	            {
	            pos++;
	            fputc('\n',output);
	            }
                 }

              MemoryFromMemory msg((long)pos,bound,this);
              pos+=msg.DevelopMail(output);
	      }
        return(pos);
        }

for (;(p=Pointer2Line(0,pos))!=NULL;pos+=length)
    {
    length=Lstrlen(p);
    fwrite(p,1,length,output);
    }
return(pos);
}
