//---------------------------------------------------------------------------
#include <limits.h>

#include "eurist.h"
//---------------------------------------------------------------------------

eurist::eurist(patch *p)
{
for (int i=0;i<p->charset_number;i++)
	score[i]=0;
factor=1;
data=p;
}

const char * eurist::charset()
{
int				index=0,i;

//printf("\n\n score [%s] = %d",Names[0],score[0]);

for (i=1;i<data->charset_number;i++)
	{
    if (score[i]>score[index])
    	index=i;
//    printf("\n score [%s] = %d",Names[i],score[i]);
    }
return(data->charset_name[index]);
}

void eurist::operator += (unsigned char *p)
{
unsigned char	c;
int		i;

for (i=0;i<data->charset_number;i++)
    if (score[i] > ULONG_MAX/2 )
    	{
        for (i=0;i<data->charset_number;i++)
        	score[i]/=2;
        factor*=2;
		break;
        }

c='\n';

do	{
	for (i=0;i<data->charset_number;i++)
	    score[i]+=data->frequencies[data->charset_table[i][c]][data->charset_table[i][*p]]/factor;
        c=*p;
	} while ((*p) && (*(p++)!='\n'));
}

#ifdef GATHER
#include <stdio.h>
#include <stdlib.h>

static unsigned int (*cmp_frequencies)[256];

static int qsort_cmp(const void *s1, const void *s2)
{
#define x ((unsigned char *)s1)
#define y ((unsigned char *)s2)

return(cmp_frequencies[x[0]][x[1]]-cmp_frequencies[y[0]][y[1]]);
}

void eurist::print_stat()
{
unsigned char (*str)[256][2];

str=new(unsigned char[256][256][2]);

printf("Most popular sequencies are\nSorting ...\r");

for (int i=0;i<256;i++)
    for (int j=0;j<256;j++)
        {
        str[i][j][0]=i;
        str[i][j][1]=j;
        }

cmp_frequencies=frequencies;
qsort(str,256*256,2,&qsort_cmp);

for (int i=234;i<256;i++)
    printf("\"%c%c\": %u \n",str[255][i][0],str[255][i][1],frequencies[str[255][i][0]][str[255][i][1]]);

delete(str);
}

int eurist::gather(char *path)
{
FILE            *fp;
int             c1,c2;

unsigned int	min=UINT_MAX,max=0;

if (data->charset_number<1)
   {
   fputs("One charset should be defined at least",stderr);
   return(-1);
   }

if ((fp=fopen(path,"rt"))==NULL)
   {
   perror("Can't open statistics file");
   return(-1);
   }

for (int i=0;i<256;i++)
	for (int j=0;j<256;j++)
        	{
                unsigned int foo;

                frequencies[i][j]=foo=(unsigned int)(data->factor*data->frequencies[i][j]);

                if (foo>max)
                   max=foo;
                if ((foo) && (foo<min))
                   min=foo;
                }

print_stat();
printf("Starting gather: begin with factor %g, min=%u, max=%u\n",data->factor, min, max);

for (int i=c1=0;(c2=getc(fp))!=EOF;i++)
    {
    c2=data->charset_table[0][c2];
    if (c1 || c2)
       {
       unsigned int foo;
    
       foo=++frequencies[c1][c2];

       if (foo>max)
          max=foo;
       c1=c2;
       }
    if ((i&0x1FFF)==0)
       printf("max=%u\r",max);
    }
fclose(fp);

print_stat();
data->factor=max/((double) USHRT_MAX );

for (int i=0;i<256;i++)
	for (int j=0;j<256;j++)
                {
                unsigned int foo;

                data->frequencies[i][j]=(unsigned short) ((foo=frequencies[i][j])/data->factor);

                if ((foo) && (foo<min))
                   min=foo;
                }

printf("Finished gather: with factor %g, min=%u, max=%u\n",data->factor, min, max);

return(0);
}

#endif	//GATHER
